@@ -1,18 +1,20 @@
-;;; ja-dic-cnv.el --- convert a Japanese dictionary (SKK-JISYO.L) to Emacs Lisp
+;;; ja-dic-cnv.el --- convert a Japanese dictionary (SKK-JISYO.L) to Emacs Lisp  -*- lexical-binding: t; -*-
+
+;; Copyright (C) 2001-2024 Free Software Foundation, Inc.
 
 ;; Copyright (C) 1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004,
-;;   2005, 2006, 2007
+;;   2005, 2006, 2007, 2008, 2009, 2010, 2011
 ;;   National Institute of Advanced Industrial Science and Technology (AIST)
 ;;   Registration Number H14PRO021
 
-;; Keywords: mule, multilingual, Japanese
+;; Keywords: i18n, mule, multilingual, Japanese
 
 ;; This file is part of GNU Emacs.
 
-;; GNU Emacs is free software; you can redistribute it and/or modify
+;; GNU Emacs is free software: you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
-;; the Free Software Foundation; either version 2, or (at your option)
-;; any later version.
+;; the Free Software Foundation, either version 3 of the License, or
+;; (at your option) any later version.
 
 ;; GNU Emacs is distributed in the hope that it will be useful,
 ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -20,9 +22,7 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to the
-;; Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-;; Boston, MA 02110-1301, USA.
+;; along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -32,95 +32,82 @@
 ;; input method (e.g. quail-japanese) can utilize the dictionary.
 
 ;; The format of SKK dictionary is quite simple.  Each line has the
-;; form "KANASTRING /CONV1/CONV2/.../" which means KANASTRING ($B2>L>J8(B
-;; $B;zNs(B) can be converted to one of CONVi.  CONVi is a Kanji ($B4A;z(B)
-;; and Kana ($B2>L>(B) mixed string.
+;; form "KANASTRING /CONV1/CONV2/.../" which means KANASTRING (仮名文
+;; 字列) can be converted to one of CONVi.  CONVi is a Kanji (漢字)
+;; and Kana (仮名) mixed string.
 ;;
-;; KANASTRING may have a trailing ASCII letter for Okurigana ($BAw$j2>L>(B)
+;; KANASTRING may have a trailing ASCII letter for Okurigana (送り仮名)
 ;; information.  For instance, the trailing letter `k' means that one
-;; of the following Okurigana is allowed: $B$+$-$/$1$3(B.  So, in that
-;; case, the string "KANASTRING$B$/(B" can be converted to one of "CONV1$B$/(B",
-;; CONV2$B$/(B, ...
+;; of the following Okurigana is allowed: かきくけこ.  So, in that
+;; case, the string "KANASTRINGく" can be converted to one of "CONV1く",
+;; CONV2く, ...
 
 ;;; Code:
 
+(require 'generate-lisp-file)
+
 ;; Name of a file to generate from SKK dictionary.
 (defvar ja-dic-filename "ja-dic.el")
 
-;; To make a generated ja-dic.el smaller.
-(make-coding-system
- 'iso-2022-7bit-short
- 2 ?J
- "Like `iso-2022-7bit' but no ASCII designation before SPC."
- '(ascii nil nil nil t t nil t)
- '((safe-charsets . t)))
-
 (defun skkdic-convert-okuri-ari (skkbuf buf)
-  (message "Processing OKURI-ARI entries ...")
+  (byte-compile-info "Processing OKURI-ARI entries" t)
   (goto-char (point-min))
-  (save-excursion
-    (set-buffer buf)
+  (with-current-buffer buf
     (insert ";; Setting okuri-ari entries.\n"
 	    "(skkdic-set-okuri-ari\n"))
   (while (not (eobp))
-    (let ((from (point))
-	  to)
-      (end-of-line)
-      (setq to (point))
+    (if (/= (following-char) ?>)
+	(let ((from (point))
+	      (to (line-end-position)))
+	  (with-current-buffer buf
+	    (insert-buffer-substring skkbuf from to)
+	    (beginning-of-line)
+	    (insert "\"")
+	    (search-forward " ")
+	    (delete-char 1)		; delete the first '/'
+	    (let ((p (point)))
+	      (end-of-line)
+	      (delete-char -1)		; delete the last '/'
+	      (subst-char-in-region p (point) ?/ ? 'noundo))
+	    (insert "\"\n"))))
 
-      (save-excursion
-	(set-buffer buf)
-	(insert-buffer-substring skkbuf from to)
-	(beginning-of-line)
-	(insert "\"")
-	(search-forward " ")
-	(delete-char 1)			; delete the first '/'
-	(let ((p (point)))
-	  (end-of-line)
-	  (delete-char -1)		; delete the last '/'
-	  (subst-char-in-region p (point) ?/ ? 'noundo))
-	(insert "\"\n"))
-
-      (forward-line 1)))
-  (save-excursion
-    (set-buffer buf)
+    (forward-line 1))
+  (with-current-buffer buf
     (insert ")\n\n")))
 
 (defconst skkdic-postfix-list '(skkdic-postfix-list))
 
 (defconst skkdic-postfix-data
-  '(("$B$$$-(B" "$B9T(B")
-    ("$B$,$+$j(B" "$B78(B")
-    ("$B$,$/(B" "$B3X(B")
-    ("$B$,$o(B" "$B@n(B")
-    ("$B$7$c(B" "$B<R(B")
-    ("$B$7$e$&(B" "$B=8(B")
-    ("$B$7$g$&(B" "$B>^(B" "$B>k(B")
-    ("$B$8$g$&(B" "$B>k(B")
-    ("$B$;$s(B" "$B@~(B")
-    ("$B$@$1(B" "$B3Y(B")
-    ("$B$A$c$/(B" "$BCe(B")
-    ("$B$F$s(B" "$BE9(B")
-    ("$B$H$&$2(B" "$BF=(B")
-    ("$B$I$*$j(B" "$BDL$j(B")
-    ("$B$d$^(B" "$B;3(B")
-    ("$B$P$7(B" "$B66(B")
-    ("$B$O$D(B" "$BH/(B")
-    ("$B$b$/(B" "$BL\(B")
-    ("$B$f$-(B" "$B9T(B")))
+  '(("いき" "行")
+    ("がかり" "係")
+    ("がく" "学")
+    ("がわ" "川")
+    ("しゃ" "社")
+    ("しゅう" "集")
+    ("しょう" "賞" "城")
+    ("じょう" "城")
+    ("せん" "線")
+    ("だけ" "岳")
+    ("ちゃく" "着")
+    ("てん" "店")
+    ("とうげ" "峠")
+    ("どおり" "通り")
+    ("やま" "山")
+    ("ばし" "橋")
+    ("はつ" "発")
+    ("もく" "目")
+    ("ゆき" "行")))
 
-(defun skkdic-convert-postfix (skkbuf buf)
-  (message "Processing POSTFIX entries ...")
+(defun skkdic-convert-postfix (_skkbuf buf)
+  (byte-compile-info "Processing POSTFIX entries" t)
   (goto-char (point-min))
-  (save-excursion
-    (set-buffer buf)
+  (with-current-buffer buf
     (insert ";; Setting postfix entries.\n"
 	    "(skkdic-set-postfix\n"))
 
   ;; Initialize SKKDIC-POSTFIX-LIST by predefined data
   ;; SKKDIC-POSTFIX-DATA.
-  (save-excursion
-    (set-buffer buf)
+  (with-current-buffer buf
     (let ((l skkdic-postfix-data)
 	  kana candidates entry)
       (while l
@@ -139,16 +126,15 @@
 	(setq l (cdr l)))))
 
   ;; Search postfix entries.
-  (while (re-search-forward "^[#<>?]\\(\\(\\cH\\|$B!<(B\\)+\\) " nil t)
-    (let ((kana (match-string 1))
+  (while (re-search-forward "^[#<>?]\\(\\cH+\\) " nil t)
+    (let ((kana (match-string-no-properties 1))
 	  str candidates)
       (while (looking-at "/[#0-9 ]*\\([^/\n]*\\)/")
-	(setq str (match-string 1))
+        (setq str (match-string-no-properties 1))
 	(if (not (member str candidates))
 	    (setq candidates (cons str candidates)))
 	(goto-char (match-end 1)))
-      (save-excursion
-	(set-buffer buf)
+      (with-current-buffer buf
 	(insert "\"" kana)
 	(while candidates
 	  (insert " " (car candidates))
@@ -161,38 +147,34 @@
 				skkdic-postfix-list)))
 	  (setq candidates (cdr candidates)))
 	(insert "\"\n"))))
-  (save-excursion
-    (set-buffer buf)
+  (with-current-buffer buf
     (insert ")\n\n")))
 
 (defconst skkdic-prefix-list '(skkdic-prefix-list))
 
-(defun skkdic-convert-prefix (skkbuf buf)
-  (message "Processing PREFIX entries ...")
+(defun skkdic-convert-prefix (_skkbuf buf)
+  (byte-compile-info "Processing PREFIX entries" t)
   (goto-char (point-min))
-  (save-excursion
-    (set-buffer buf)
+  (with-current-buffer buf
     (insert ";; Setting prefix entries.\n"
 	    "(skkdic-set-prefix\n"))
   (save-excursion
-    (while (re-search-forward "^\\(\\(\\cH\\|$B!<(B\\)+\\)[<>?] " nil t)
-      (let ((kana (match-string 1))
+    (while (re-search-forward "^\\(\\cH+\\)[<>?] " nil t)
+      (let ((kana (match-string-no-properties 1))
 	    str candidates)
 	(while (looking-at "/\\([^/\n]+\\)/")
-	  (setq str (match-string 1))
+          (setq str (match-string-no-properties 1))
 	  (if (not (member str candidates))
 	      (setq candidates (cons str candidates)))
 	  (goto-char (match-end 1)))
-	(save-excursion
-	  (set-buffer buf)
+	(with-current-buffer buf
 	  (insert "\"" kana)
 	  (while candidates
 	    (insert " " (car candidates))
 	    (set-nested-alist (car candidates) kana skkdic-prefix-list)
 	    (setq candidates (cdr candidates)))
 	  (insert "\"\n")))))
-  (save-excursion
-    (set-buffer buf)
+  (with-current-buffer buf
     (insert ")\n\n")))
 
 ;; FROM and TO point the head and tail of "/J../J../.../".
@@ -200,8 +182,8 @@
   (let (candidates)
     (goto-char from)
     (while (re-search-forward "/[^/ \n]+" to t)
-      (setq candidates (cons (buffer-substring (1+ (match-beginning 0))
-					       (match-end 0))
+      (setq candidates (cons (buffer-substring-no-properties
+                              (1+ (match-beginning 0)) (match-end 0))
 			     candidates)))
     candidates))
 
@@ -215,8 +197,8 @@
 ;; Return t if substring of STR (between FROM and TO) can be broken up
 ;; to chunks all of which can be derived from another entry in SKK
 ;; dictionary.  SKKBUF is the buffer where the original SKK dictionary
-;; is visited, KANA is the current entry for STR.  FIRST is t iff this
-;; is called at top level.
+;; is visited, KANA is the current entry for STR.  FIRST is t only if
+;; this is called at top level.
 
 (defun skkdic-breakup-string (skkbuf kana str from to &optional first)
   (let ((len (- to from)))
@@ -229,7 +211,7 @@
 				    (substring str from idx)
 				    skkdic-word-list)))
 		   (if (or (and (consp kana2-list)
-				(let ((kana-len (length kana))
+				(let (;; (kana-len (length kana))
 				      kana2)
 				  (catch 'skkdic-tag
 				    (while kana2-list
@@ -271,12 +253,16 @@
 ;; Return list of candidates which excludes some from CANDIDATES.
 ;; Excluded candidates can be derived from another entry.
 
+(defconst skkdic--japanese-category-set (make-category-set "j"))
+
 (defun skkdic-reduced-candidates (skkbuf kana candidates)
   (let (elt l)
     (while candidates
       (setq elt (car candidates))
       (if (or (= (length elt) 1)
-	      (and (string-match "^\\cj" elt)
+	      (and (bool-vector-subsetp
+                    skkdic--japanese-category-set
+                    (char-category-set (aref elt 0)))
 		   (not (skkdic-breakup-string skkbuf kana elt 0 (length elt)
 					       'first))))
 	  (setq l (cons elt l)))
@@ -287,24 +273,19 @@
 (defvar skkdic-okuri-nasi-entries-count 0)
 
 (defun skkdic-collect-okuri-nasi ()
-  (message "Collecting OKURI-NASI entries ...")
   (save-excursion
-    (let ((prev-ratio 0)
-	  ratio)
-      (while (re-search-forward "^\\(\\(\\cH\\|$B!<(B\\)+\\) \\(/\\cj.*\\)/$"
+    (let ((progress (make-progress-reporter
+                     (byte-compile-info "Collecting OKURI-NASI entries" t)
+                     (point) (point-max)
+                     nil 10)))
+      (while (re-search-forward "^\\(\\cH+\\) \\(/\\cj.*\\)/$"
 				nil t)
-	(let ((kana (match-string 1))
-	      (candidates (skkdic-get-candidate-list (match-beginning 3)
-						     (match-end 3))))
+        (let ((kana (match-string-no-properties 1))
+	      (candidates (skkdic-get-candidate-list (match-beginning 2)
+						     (match-end 2))))
 	  (setq skkdic-okuri-nasi-entries
-		(cons (cons kana candidates) skkdic-okuri-nasi-entries)
-		skkdic-okuri-nasi-entries-count
-		(1+ skkdic-okuri-nasi-entries-count))
-	  (setq ratio (floor (/ (* (point) 100.0) (point-max))))
-	  (if (/= ratio prev-ratio)
-	      (progn
-		(message "collected %2d%% %s ..." ratio kana)
-		(setq prev-ratio ratio)))
+		(cons (cons kana candidates) skkdic-okuri-nasi-entries))
+          (progress-reporter-update progress (point))
 	  (while candidates
 	    (let ((entry (lookup-nested-alist (car candidates)
 					      skkdic-word-list nil nil t)))
@@ -312,67 +293,66 @@
 		  (setcar entry (cons kana (car entry)))
 		(set-nested-alist (car candidates) (list kana)
 				  skkdic-word-list)))
-	    (setq candidates (cdr candidates))))))))
+            (setq candidates (cdr candidates)))))
+      (setq skkdic-okuri-nasi-entries-count (length skkdic-okuri-nasi-entries))
+      (progress-reporter-done progress))))
 
-(defun skkdic-convert-okuri-nasi (skkbuf buf)
-  (message "Processing OKURI-NASI entries ...")
-  (save-excursion
-    (set-buffer buf)
+(defun skkdic-convert-okuri-nasi (skkbuf buf &optional no-reduction)
+  (with-current-buffer buf
     (insert ";; Setting okuri-nasi entries.\n"
 	    "(skkdic-set-okuri-nasi\n")
     (let ((l (nreverse skkdic-okuri-nasi-entries))
-	  (count 0)
-	  (prev-ratio 0)
-	  ratio)
+          (progress (make-progress-reporter
+                     (byte-compile-info "Processing OKURI-NASI entries" t)
+                     0 skkdic-okuri-nasi-entries-count
+                     nil 10))
+          (count 0))
       (while l
 	(let ((kana (car (car l)))
 	      (candidates (cdr (car l))))
-	  (setq ratio (/ (* count 1000) skkdic-okuri-nasi-entries-count)
-		count (1+ count))
-	  (if (/= prev-ratio (/ ratio 10))
-	      (progn
-		(message "processed %2d%% %s ..." (/ ratio 10) kana)
-		(setq prev-ratio (/ ratio 10))))
+          (setq count (1+ count))
+          (progress-reporter-update progress count)
 	  (if (setq candidates
-		    (skkdic-reduced-candidates skkbuf kana candidates))
+                    (if no-reduction
+                        candidates
+                      (skkdic-reduced-candidates skkbuf kana candidates)))
 	      (progn
 		(insert "\"" kana)
 		(while candidates
 		  (insert " " (car candidates))
 		  (setq candidates (cdr candidates)))
 		(insert "\"\n"))))
-	(setq l (cdr l))))
+	(setq l (cdr l)))
+      (progress-reporter-done progress))
     (insert ")\n\n")))
 
-(defun skkdic-convert (filename &optional dirname)
-  "Generate Emacs Lisp file form Japanese dictionary file FILENAME.
+(defun skkdic-convert (filename &optional dirname no-reduction)
+  "Generate Emacs Lisp file from Japanese dictionary file FILENAME.
 The format of the dictionary file should be the same as SKK dictionaries.
-Optional argument DIRNAME if specified is the directory name under which
-the generated Emacs Lisp is saved.
-The name of generated file is specified by the variable `ja-dic-filename'."
+Saves the output as `ja-dic-filename', in directory DIRNAME (if specified).
+If NO-REDUCTION is non-nil, do not reduce the dictionary vocabulary."
   (interactive "FSKK dictionary file: ")
-  (message "Reading file \"%s\" ..." filename)
-  (let* ((coding-system-for-read 'euc-japan)
-	 (skkbuf(find-file-noselect (expand-file-name filename)))
+  (let* ((skkbuf (get-buffer-create " *skkdic-unannotated*"))
 	 (buf (get-buffer-create "*skkdic-work*")))
-    (save-excursion
-      ;; Setup and generate the header part of working buffer.
-      (set-buffer buf)
+    ;; Set skkbuf to an unannotated copy of the dictionary.
+    (with-current-buffer skkbuf
+      (let ((coding-system-for-read 'euc-japan))
+        (insert-file-contents (expand-file-name filename)))
+      (re-search-forward "^[^;]")
+      (while (re-search-forward ";[^\n/]*/" nil t)
+	(replace-match "/" t t)))
+    ;; Setup and generate the header part of working buffer.
+    (with-current-buffer buf
       (erase-buffer)
       (buffer-disable-undo)
-      (insert ";;; ja-dic.el --- dictionary for Japanese input method"
-	      " -*-coding: iso-2022-jp; byte-compile-disable-print-circle:t; -*-\n"
-	      ";;\tGenerated by the command `skkdic-convert'\n"
-	      ";;\tDate: " (current-time-string) "\n"
-	      ";;\tOriginal SKK dictionary file: "
+      (generate-lisp-file-heading ja-dic-filename 'skkdic-convert :code nil)
+      (insert (format ";; Generated with small ja-dic option: %s\n\n"
+                      (if no-reduction "no" "yes")))
+      (insert ";; Original SKK dictionary file: "
 	      (file-relative-name (expand-file-name filename) dirname)
 	      "\n\n"
-	      ";; This file is part of GNU Emacs.\n\n"
-	      ";;; Commentary:\n\n"
-	      ";; Do byte-compile this file again after any modification.\n\n"
 	      ";;; Start of the header of the original SKK dictionary.\n\n")
       (set-buffer skkbuf)
-      (widen)
       (goto-char 1)
       (let (pos)
 	(search-forward ";; okuri-ari")
@@ -413,18 +393,17 @@ The name of generated file is specified by the variable `ja-dic-filename'."
 	(skkdic-collect-okuri-nasi)
 
 	;; Convert okuri-nasi general entries.
-	(skkdic-convert-okuri-nasi skkbuf buf)
+	(skkdic-convert-okuri-nasi skkbuf buf no-reduction)
 
 	;; Postfix
-	(save-excursion
-	  (set-buffer buf)
+	(with-current-buffer buf
 	  (goto-char (point-max))
-	  (insert ";;\n(provide 'ja-dic)\n\n;;; ja-dic.el ends here\n")))
+          (generate-lisp-file-trailer ja-dic-filename :compile t)))
 
       ;; Save the working buffer.
       (set-buffer buf)
       (set-visited-file-name (expand-file-name ja-dic-filename dirname) t)
-      (set-buffer-file-coding-system 'iso-2022-7bit-short)
+      (set-buffer-file-coding-system 'utf-8)
       (save-buffer 0))
     (kill-buffer skkbuf)
     (switch-to-buffer buf)))
@@ -446,20 +425,21 @@ To get complete usage, invoke:
 	(message "To convert SKK-JISYO.L into skkdic.el:")
 	(message "  %% emacs -batch -l ja-dic-cnv -f batch-skkdic-convert SKK-JISYO.L")
 	(message "To convert SKK-JISYO.L into DIR/ja-dic.el:")
-	(message "  %% emacs -batch -l ja-dic-cnv -f batch-skkdic-convert -dir DIR SKK-JISYO.L"))
-    (let (targetdir filename)
+	(message "  %% emacs -batch -l ja-dic-cnv -f batch-skkdic-convert -dir DIR SKK-JISYO.L")
+        (message "To convert SKK-JISYO.L into skkdic.el without reducing dictionary vocabulary:")
+        (message "  %% emacs -batch -l ja-dic-cnv -f batch-skkdic-convert --no-reduction SKK-JISYO.L"))
+    (let (targetdir filename no-reduction)
       (if (string= (car command-line-args-left) "-dir")
 	  (progn
 	    (setq command-line-args-left (cdr command-line-args-left))
 	    (setq targetdir (expand-file-name (car command-line-args-left)))
 	    (setq command-line-args-left (cdr command-line-args-left))))
+      (if (string= (car command-line-args-left) "--no-reduction")
+          (progn
+	    (setq no-reduction t)
+	    (setq command-line-args-left (cdr command-line-args-left))))
       (setq filename (expand-file-name (car command-line-args-left)))
-      (message "Converting %s to %s ..." filename ja-dic-filename)
-      (message "It takes around 10 minutes even on Sun SS20.")
-      (skkdic-convert filename targetdir)
-      (message "Do byte-compile the created file by:")
-      (message "  %% emacs -batch -f batch-byte-compile %s" ja-dic-filename)
-      ))
+      (skkdic-convert filename targetdir no-reduction)))
   (kill-emacs 0))
 
 
@@ -476,9 +456,9 @@ To get complete usage, invoke:
       (aset vec i
 	    (if (< ch 128)		; CH is an ASCII letter for OKURIGANA,
 		(- ch)			;  represented by a negative code.
-	      (if (= ch ?$B!<(B)		; `$B!<(B' is represented by 0.
+	      (if (= ch ?ー)		; `ー' is represented by 0.
 		  0
-		(- (nth 2 (split-char ch)) 32))))
+		(- (logand (encode-char ch 'japanese-jisx0208) #xFF) 32))))
       (setq i (1+ i)))
     vec))
 
@@ -488,7 +468,7 @@ To get complete usage, invoke:
 	(i (match-end 0))
 	candidates)
     (while (string-match "[^ ]+" entry i)
-      (setq candidates (cons (match-string 0 entry) candidates))
+      (setq candidates (cons (match-string-no-properties 0 entry) candidates))
       (setq i (match-end 0)))
     (cons (skkdic-get-kana-compact-codes kana) candidates)))
 
@@ -553,22 +533,18 @@ To get complete usage, invoke:
   `(defconst skkdic-okuri-nasi
      ',(let ((l entries)
 	     (map '(skdic-okuri-nasi))
+             (progress (make-progress-reporter
+                        (byte-compile-info "Extracting OKURI-NASI entries")
+                        0 (length entries)))
 	     (count 0)
 	     entry)
 	 (while l
-	   (setq count (1+ count))
-	   (if (= (% count 10000) 0)
-	       (message (format "%d entries" count)))
+           (progress-reporter-update progress (setq count (1+ count)))
 	   (setq entry (skkdic-extract-conversion-data (car l)))
 	   (set-nested-alist (car entry) (cdr entry) map)
 	   (setq l (cdr l)))
+         (progress-reporter-done progress)
 	 map)))
 
 (provide 'ja-dic-cnv)
-
-;; Local Variables:
-;; coding: iso-2022-7bit
-;; End:
-
-;;; arch-tag: dec06fb0-8118-45b1-80d7-dc360b6fd3b2
 ;;; ja-dic-cnv.el ends here
